#include "SAB.h"

SAB::SAB() : _screenConfig(128, 64, 0x3C),
_sdCardConfig(SPI_FULL_SPEED), 
_display(_screenConfig.getWidth(), 
_screenConfig.getHeight(), &Wire), 
_screenManager(_display), _rtc(), 
_rtcManager(_rtc), _sdCardManager(), 
_connectivityManager(NULL), 
_webServerManager(80, &_sdCardManager), 
_error(0)
{ 
  Serial.begin(9600);
  delay(200);
  
  //We initialize the pins for the  I2C communication
  Wire.begin(_pinConfig.getI2C_sda(), _pinConfig.getI2C_scl());
  if(!_rtc.begin()) _error |= RTC_BEGIN_ERR;
  if(!_display.begin(SSD1306_SWITCHCAPVCC, _screenConfig.getAddress())){ _error |= DISP_BEGIN_ERR; }
  if(!_sdCardManager.begin(_sdCardConfig.getSPISpeed(), _pinConfig.getSPI_sdCard_cs())){ _error |= SDCARD_INIT_ERR; }
  _connectivityManager = new ConnectivityManager(&_sdCardManager);
}

SAB::SAB(const PinMapping pinConfig, const ScreenConfig screenConfig, const SDCardConfig sdCardConfig, const unsigned int webServerPort) : _pinConfig(pinConfig), 
_screenConfig(screenConfig), 
_sdCardConfig(sdCardConfig), 
_display(_screenConfig.getWidth(), _screenConfig.getHeight(), &Wire), 
_screenManager(_display), 
_rtc(), 
_rtcManager(_rtc), 
_sdCardManager(), 
_connectivityManager(NULL), 
_webServerManager(webServerPort, &_sdCardManager), 
_error(0)
{
  Serial.begin(9600);
  delay(200);
  
  //We initialize the pins for the  I2C communication
  Wire.begin(_pinConfig.getI2C_sda(), _pinConfig.getI2C_scl());
  if(!_rtc.begin()) _error |= RTC_BEGIN_ERR;
  if(!_display.begin(SSD1306_SWITCHCAPVCC, _screenConfig.getAddress())) _error |= DISP_BEGIN_ERR;
  if(!_sdCardManager.begin(_sdCardConfig.getSPISpeed(), _pinConfig.getSPI_sdCard_cs())){ _error |= SDCARD_INIT_ERR;}
  _connectivityManager = new ConnectivityManager(&_sdCardManager);
}

ScreenManager& SAB::getScreenManager()
{
  return _screenManager;
}

RtcManager& SAB::getRtcManager()
{
  return _rtcManager;
}

SDCardManager& SAB::getSdCardManager()
{
  return _sdCardManager;
}

ConnectivityManager& SAB::getConnectivityManager()
{
  return *_connectivityManager;
}

WEBServerManager& SAB::getWebServerManager()
{
  return _webServerManager;
}

ScreenConfig SAB::getScreenConfig() const
{
  return _screenConfig;
}

PinMapping SAB::getPinConfig() const
{
  return _pinConfig;
}

SDCardConfig SAB::getSdCardConfig() const
{
  return _sdCardConfig;
}

PowerInfo SAB::getPowerInfo(const unsigned char nbOfMeasures) const
{
  uint16_t batteryMeasures = 0, temp = 0;
  unsigned char i = 0;
  int checkLevel = 0;
  PowerInfo p;

  for(;i < nbOfMeasures;i++)
  {
    temp = analogRead(ADC);
    batteryMeasures += temp;
  }
  
  batteryMeasures /= nbOfMeasures;
  
  if(batteryMeasures <= USB_THRESHOLD)
  {
    p.powerType = USB;
    p.level = 0;
  }
  else 
  {
    p.powerType = BATTERY; 
    checkLevel = ((batteryMeasures-BATT_EMPTY)*100/BATT_DIFF);
    
    if(checkLevel > 100)
      p.level = 100;
    else if (checkLevel < 0)
      p.level = 0;
    else
      p.level = checkLevel;
  }
  
  return p;
}

unsigned char SAB::getError() const
{
  return _error;
}

